# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
@software{jax2018github,
  author = {James Bradbury and Roy Frostig and Peter Hawkins and Matthew James Johnson and Chris Leary and Dougal Maclaurin and George Necula and Adam Paszke and Jake Vander{P}las and Skye Wanderman-{M}ilne and Qiao Zhang},
  title = {{JAX}: composable transformations of {P}ython+{N}um{P}y programs},
  url = {http://github.com/google/jax},
  version = {0.2.5},
  year = {2018},
}
Datasets used in examples.
"""


import array
import gzip
import os
from os import path
import struct
import urllib.request

import numpy as np


_DATA = "/opt/ml/tmp/jax_example_data/"


def _download(url, filename):
    """Download a url to a file in the JAX data temp directory."""
    if not path.exists(_DATA):
        os.makedirs(_DATA)
    out_file = path.join(_DATA, filename)
    if not path.isfile(out_file):
        urllib.request.urlretrieve(url, out_file)
        print("downloaded {} to {}".format(url, _DATA))


def _partial_flatten(x):
    """Flatten all but the first dimension of an ndarray."""
    return np.reshape(x, (x.shape[0], -1))


def _one_hot(x, k, dtype=np.float32):
    """Create a one-hot encoding of x of size k."""
    return np.array(x[:, None] == np.arange(k), dtype)


def mnist_raw():
    """Download and parse the raw MNIST dataset."""
    # CVDF mirror of http://yann.lecun.com/exdb/mnist/
    base_url = "https://storage.googleapis.com/cvdf-datasets/mnist/"

    def parse_labels(filename):
        with gzip.open(filename, "rb") as fh:
            _ = struct.unpack(">II", fh.read(8))
            return np.array(array.array("B", fh.read()), dtype=np.uint8)

    def parse_images(filename):
        with gzip.open(filename, "rb") as fh:
            _, num_data, rows, cols = struct.unpack(">IIII", fh.read(16))
            return np.array(array.array("B", fh.read()), dtype=np.uint8).reshape(
                num_data, rows, cols
            )

    for filename in [
        "train-images-idx3-ubyte.gz",
        "train-labels-idx1-ubyte.gz",
        "t10k-images-idx3-ubyte.gz",
        "t10k-labels-idx1-ubyte.gz",
    ]:
        _download(base_url + filename, filename)

    train_images = parse_images(path.join(_DATA, "train-images-idx3-ubyte.gz"))
    train_labels = parse_labels(path.join(_DATA, "train-labels-idx1-ubyte.gz"))
    test_images = parse_images(path.join(_DATA, "t10k-images-idx3-ubyte.gz"))
    test_labels = parse_labels(path.join(_DATA, "t10k-labels-idx1-ubyte.gz"))

    return train_images, train_labels, test_images, test_labels


def mnist(permute_train=False):
    """Download, parse and process MNIST data to unit scale and one-hot labels."""

    train_images, train_labels, test_images, test_labels = mnist_raw()

    train_images = _partial_flatten(train_images) / np.float32(255.0)
    test_images = _partial_flatten(test_images) / np.float32(255.0)
    train_labels = _one_hot(train_labels, 10)
    test_labels = _one_hot(test_labels, 10)

    if permute_train:
        perm = np.random.RandomState(0).permutation(train_images.shape[0])
        train_images = train_images[perm]
        train_labels = train_labels[perm]

    return train_images, train_labels, test_images, test_labels



from jax import lax
from jax import numpy as np
from jax import tree_util
from jax import jit, random
from datetime import datetime

@jit
def pytree_dot(x, y) -> float:
    partial_dot = tree_util.tree_multimap(lambda arr1, arr2: np.sum(arr1 * arr2), x, y)
    return tree_util.tree_reduce(lax.add, partial_dot)


def pytree_dot2(x, y) -> float:
    partial_dot = tree_util.tree_multimap(lambda arr1, arr2: np.sum(arr1 * arr2), x, y)
    return tree_util.tree_reduce(lax.add, partial_dot)

@jit
def _nested_jit(x, y):
    s = pytree_dot(x, y)
    s += pytree_dot(x, y)
    s += pytree_dot(x, y)
    s += pytree_dot(x, y)
    s += pytree_dot(x, y)
    return s

@jit
def _jit(x, y):
    s = pytree_dot2(x, y)
    s += pytree_dot2(x, y)
    s += pytree_dot2(x, y)
    s += pytree_dot2(x, y)
    s += pytree_dot2(x, y)
    return s


def no_jit(x, y):
    s = pytree_dot2(x, y)
    s += pytree_dot2(x, y)
    s += pytree_dot2(x, y)
    s += pytree_dot2(x, y)
    s += pytree_dot2(x, y)
    return s


def internal_only_jit(x, y):
    s = pytree_dot(x, y)
    s += pytree_dot(x, y)
    s += pytree_dot(x, y)
    s += pytree_dot(x, y)
    s += pytree_dot(x, y)
    return s

if __name__ == '__main__':
    k1, k2, k3 = random.split(random.PRNGKey(3), 3)
    x = [np.array([0.5, 0.5]), random.uniform(k1, shape=(50000,)), {"beta": 0.5}]
    y = [np.array([0.5, 0.5]), random.uniform(k2, shape=(50000,)), {"beta": 0.5}]
    z = [np.array([0.5, 0.5]), random.uniform(k3, shape=(50000,)), {"beta": 0.5}]
    x1 = x
    s = 0.0

    start_time = datetime.now()
    for i in range(10000):
        if (i%2 ==0):
            x1 = z
        else:
            x1 = x
        s = _jit(x, y)
    end_time = datetime.now()
    print('Std Jit Duration: {}'.format(end_time - start_time))
    print(f'Dot: {s}')

    start_time = datetime.now()
    for i in range(100000):
        if (i % 2 == 0):
            x1 = z
        else:
            x1 = x
        s = _nested_jit(x, y)

    end_time = datetime.now()
    print('Nested Jit Duration: {}'.format(end_time - start_time))
    print(f'Dot: {s}')

    start_time = datetime.now()
    for i in range(100000):
        if (i % 2 == 0):
            x1 = z
        else:
            x1 = x
        s = internal_only_jit(x, y)

    end_time = datetime.now()
    print('internal ops jit Duration: {}'.format(end_time - start_time))
    print(f'Dot: {s}')

    start_time = datetime.now()
    for i in range(100000):
        if (i % 2 == 0):
            x1 = z
        else:
            x1 = x
        s = _nested_jit(x, y)

    end_time = datetime.now()
    print('No Jit Duration: {}'.format(end_time - start_time))








from abc import ABC, abstractmethod


class Continuation(ABC):
    """Abstract Continuation Strategy to be inherited by developer for any new Continuation Strategy."""

    @abstractmethod
    def run(self):
        pass



from cjax.continuation.methods.corrector.constrained_corrector import (
    ConstrainedCorrector,
)
from typing import Tuple
from jax.tree_util import tree_map
from jax import random
from cjax.utils.rotation_ndims import *
from jax import jit
import math


class PerturbedCorrecter(ConstrainedCorrector):
    """Minimize the objective using gradient based method along with some constraint and noise"""

    def __init__(
        self,
        optimizer,
        objective,
        dual_objective,
        lagrange_multiplier,
        concat_states,
        delta_s,
        ascent_opt,
        key_state,
        compute_min_grad_fn,
        compute_max_grad_fn,
        compute_grad_fn,
        hparams,
        pred_state,
        pred_prev_state,
        counter
    ):
        super().__init__(
            optimizer,
            objective,
            dual_objective,
            lagrange_multiplier,
            concat_states,
            delta_s,
            ascent_opt,
            compute_min_grad_fn,
            compute_max_grad_fn,
            compute_grad_fn,
            hparams,
        )
        self._parc_vec = None
        self.state_stack = dict()
        _, self.key = random.split(random.PRNGKey(key_state))
        self.pred_state = pred_state
        self.pred_prev_state = pred_prev_state
        self.sphere_radius = hparams["sphere_radius"]
        self.counter = counter

    @staticmethod
    @jit
    def exp_decay(epoch, initial_lrate):
        k = 0.1
        lrate = initial_lrate * np.exp(-k * epoch)
        return lrate

    @staticmethod
    @jit
    def _perform_perturb_by_projection(
            _state_secant_vector,
            _state_secant_c2,
            key,
            pred_prev_state,
            _state,
            _bparam,
            counter,
            sphere_radius,
    ):
        ### Secant normal
        n, sample_unravel = pytree_to_vec(
            [_state_secant_vector["state"], _state_secant_vector["bparam"]]
        )
        n = pytree_normalized(n)
        ### sample a random poin in Rn
        u = tree_map(
            lambda a: a + random.uniform(key, a.shape),
            pytree_zeros_like(n),
        )

        tmp, _ = pytree_to_vec([_state_secant_c2['state'], _state_secant_c2['bparam']])

        # select a point on the secant normal
        u_0, _ = pytree_to_vec(pred_prev_state)
        # compute projection
        proj_of_u_on_n = projection_affine(len(n), u, n, u_0)

        point_on_plane = u + pytree_sub(tmp, proj_of_u_on_n)  ## state= pred_state + n
        inv_vec = np.array([-1.0, 1.0])
        parc = pytree_element_mul(
            pytree_normalized(pytree_sub(point_on_plane, tmp)),
            inv_vec[(counter % 2)],
        )
        point_on_plane_2 = tmp + sphere_radius * parc
        new_sample = sample_unravel(point_on_plane_2)
        state_stack = {}
        state_stack.update({"state": new_sample[0]})
        state_stack.update({"bparam": new_sample[1]})
        _parc_vec = pytree_sub(state_stack, _state_secant_c2)
        return _parc_vec, state_stack

    def _evaluate_perturb(self):
        """Evaluate weather the perturbed vector is orthogonal to secant vector"""

        dot = pytree_dot(pytree_normalized(self._parc_vec), pytree_normalized(self._state_secant_vector))
        if math.isclose(dot, 0.0, abs_tol=0.25):
            print(f"Perturb was near arc-plane. {dot}")
            self._state = self.state_stack["state"]
            self._bparam = self.state_stack["bparam"]
        else:
            print(f"Perturb was not on arc-plane.{dot}")

    def correction_step(self) -> Tuple:
        """Given the current state optimize to the correct state.

        Returns:
          (state: problem parameters, bparam: continuation parameter) Tuple
        """
        self._parc_vec, self.state_stack = self._perform_perturb_by_projection(
            self._state_secant_vector,
            self._state_secant_c2,
            self.key,
            self.pred_prev_state,
            self._state,
            self._bparam,
            self.counter,
            self.sphere_radius,
        )
        self._evaluate_perturb()  # does every time
        quality =0.0
        for k in range(self.warmup_period):
            grads = self.compute_grad_fn(self._state, self._bparam)
            self._state = self.opt.update_params(self._state, grads[0])

        for k in range(self.ascent_period):
            lagrange_grads = self.compute_max_grad_fn(
                self._state,
                self._bparam,
                self._lagrange_multiplier,
                self._state_secant_c2,
                self._state_secant_vector,
                self.delta_s,
            )
            self._lagrange_multiplier = self.ascent_opt.update_params(
                self._lagrange_multiplier, lagrange_grads[0], k
            )
            for j in range(self.descent_period):
                state_grads, bparam_grads = self.compute_min_grad_fn(
                    self._state,
                    self._bparam,
                    self._lagrange_multiplier,
                    self._state_secant_c2,
                    self._state_secant_vector,
                    self.delta_s,
                )

                if self.hparams['adaptive']:
                    self.opt.lr = self.exp_decay(j, self.hparams['natural_lr'])
                    quality = l2_norm(state_grads) + l2_norm(bparam_grads)
                    # if quality > self.hparams['quality_thresh']:
                    #     self.hparams['natural_lr'] = int(self.hparams['natural_lr']) / 8
                        # print('grads', bparam_grads, state_grads)
                        # state_grads = clip_grads(state_grads, 0.01)
                        # bparam_grads = clip_grads(bparam_grads, 0.01)
                        # print('clipped grads', bparam_grads, state_grads)

                self._bparam = self.opt.update_params(self._bparam, bparam_grads, j)
                self._state = self.opt.update_params(self._state, state_grads, j)

        return self._state, self._bparam, quality



from jax.nn.initializers import zeros
from jax.nn import sigmoid
from jax.experimental.optimizers import l2_norm
from jax.experimental.stax import Dense, elementwise, Identity
from jax import random
import jax.numpy as np


def constant_2d(m, dtype=np.float32):
    """
    Constant weight initializer
    Args:
        m: constant matrix
        dtype: matrix dtype

    Returns:
        init function
    """

    def init(key, shape, dtype=dtype):
        return m[: shape[0], : shape[1]]

    return init


def HomotopyDropout(rate, mode="train"):
    """Layer construction function for a dropout layer with given rate."""

    def init_fun(rng, input_shape):
        return input_shape, ()

    def apply_fun(params, inputs, **kwargs):
        rng = kwargs.get("rng", None)
        if rng is None:
            msg = (
                "Dropout layer requires apply_fun to be called with a PRNG key "
                "argument. That is, instead of `apply_fun(params, inputs)`, call "
                "it like `apply_fun(params, inputs, rng)` where `rng` is a "
                "jax.random.PRNGKey value."
            )
            raise ValueError(msg)
        if mode == "train":
            hrate = kwargs.get("bparam") + rate
            _, key = random.split(random.PRNGKey(0))
            keep = random.bernoulli(key, hrate, inputs.shape)
            return np.where(keep, inputs, 0)
        else:
            return inputs

    return init_fun, apply_fun


def v_2d(m, dtype=np.float32):
    """
    Constant weight initializer
    Args:
        m: constant matrix
        dtype: matrix dtype

    Returns:
        init function
    """

    def init(key, shape, dtype=dtype):
        return m[: shape[0], : shape[1]]

    return init


def constant(matrix, dtype=np.float32):
    def init(key, shape, dtype=dtype):
        if matrix.shape != shape:
            raise ValueError(
                f"input array of shape {matrix.shape} is not equal to weight array of shape {shape}"
            )
        return matrix

    return init


def homotopy_activation(x, alpha=1.0, activation_func=None):
    """

    Args:
        x:
        alpha:
        activation_func:

    Returns:

    """
    if activation_func:
        return alpha * x + (1 - alpha) * activation_func(x)
    else:
        return x


def HomotopyDense(out_dim, W_init=zeros, b_init=zeros):
    """
    Layer constructor function for a dense (fully-connected) layer.
    Args:
        out_dim:
        W_init:
        b_init:

    Returns:

    """

    def init_fun(rng, input_shape):
        output_shape = input_shape[:-1] + (out_dim,)
        k1, k2 = random.split(rng)
        W, b = W_init(k1, (input_shape[-1], out_dim)), b_init(k2, (out_dim,))
        return output_shape, (W, b)

    def apply_fun(params, inputs, **kwargs):
        bparam = kwargs.get("bparam")
        W, b = params
        dense = np.dot(inputs, W) + b
        return homotopy_activation(
            dense, alpha=bparam, activation_func=kwargs.get("activation_func")
        )

    return init_fun, apply_fun


def LambdaIdentity():
    """Layer construction function for an identity layer."""
    init_fun = lambda rng, input_shape: (input_shape, ())
    apply_fun = lambda params, inputs, **kwargs: inputs * kwargs.get("bparam")
    return init_fun, apply_fun


LambdaIdentity = LambdaIdentity()



from cjax.continuation.base_continuation import Continuation
from cjax.continuation.states.state_variables import StateVariable, StateWriter
from cjax.optimizer.optimizer import OptimizerCreator
from cjax.continuation.methods.predictor.natural_predictor import NaturalPredictor
from cjax.continuation.methods.corrector.unconstrained_corrector import (
    UnconstrainedCorrector,
)
from jax.tree_util import *
import gc
from cjax.utils import profile
from jax import jit, grad
import jax.numpy as np


class NaturalContinuation(Continuation):
    """Natural Continuation strategy.

    Composed of natural predictor and unconstrained corrector"""

    def __init__(self, state, bparam, counter, objective, hparams):
        self._state_wrap = StateVariable(state, counter)
        self._bparam_wrap = StateVariable(bparam, counter)
        self.objective = objective
        self.value_func = jit(self.objective)
        self._value_wrap = StateVariable(self.objective(state, bparam), counter)
        self.sw = None
        self.hparams = hparams
        self.opt = OptimizerCreator(
            opt_string=hparams["meta"]["optimizer"], learning_rate=hparams["natural_lr"]
        ).get_optimizer()
        self.continuation_steps = hparams["continuation_steps"]

        self.output_file = hparams["meta"]["output_dir"]
        self._delta_s = hparams["delta_bparams"]
        self.grad_fn = jit(grad(self.objective, argnums=[0]))

    @profile(sort_by="cumulative", lines_to_print=10, strip_dirs=True)
    def run(self):
        """Runs the continuation strategy.

        A continuation strategy that defines how predictor and corrector components of the algorithm
        interact with the states of the mathematical system.
        """
        self.sw = StateWriter(f"{self.output_file}/version.json")

        for i in range(self.continuation_steps):
            print(self._value_wrap.get_record(), self._bparam_wrap.get_record())
            self._state_wrap.counter = i
            self._bparam_wrap.counter = i
            self._value_wrap.counter = i
            self.sw.write(
                [
                    self._state_wrap.get_record(),
                    self._bparam_wrap.get_record(),
                    self._value_wrap.get_record(),
                ]
            )

            concat_states = [self._state_wrap.state, self._bparam_wrap.state]
            predictor = NaturalPredictor(
                concat_states=concat_states, delta_s=self._delta_s
            )
            predictor.prediction_step()

            concat_states = [predictor.state, predictor.bparam]
            del predictor
            gc.collect()
            corrector = UnconstrainedCorrector(
                optimizer=self.opt,
                objective=self.objective,
                concat_states=concat_states,
                grad_fn=self.grad_fn,
                warmup_period=self.hparams["warmup_period"],
            )
            state, bparam = corrector.correction_step()

            clip_lambda = lambda g: np.where(
                (g > self.hparams["lambda_max"]), self.hparams["lambda_max"], g
            )
            bparam = tree_map(clip_lambda, bparam)
            clip_lambda = lambda g: np.where(
                (g < self.hparams["lambda_min"]), self.hparams["lambda_min"], g
            )
            bparam = tree_map(clip_lambda, bparam)

            value = self.value_func(state, bparam)
            self._state_wrap.state = state
            self._bparam_wrap.state = bparam
            self._value_wrap.state = value
            del corrector
            gc.collect()
            if self._bparam_wrap.state[0] >= self.hparams["lambda_max"]:
                break



import matplotlib.pyplot as plt
import jsonlines
import numpy as onp
from cjax.utils.math_trees import *
from typing import Tuple

"""
cmaps['Sequential'] = [
            'Greys', 'Purples', 'Blues', 'Greens', 'Oranges', 'Reds',
            'YlOrBr', 'YlOrRd', 'OrRd', 'PuRd', 'RdPu', 'BuPu',
            'GnBu', 'PuBu', 'YlGnBu', 'PuBuGn', 'BuGn', 'YlGn']
            

cmaps['Diverging'] = [
            'PiYG', 'PRGn', 'BrBG', 'PuOr', 'RdGy', 'RdBu',
            'RdYlBu', 'RdYlGn', 'Spectral', 'coolwarm', 'bwr', 'seismic']
            
plt.scatter(y, x, c=z, cmap='gray_r')
"""


def pick_array(data: list, start: int = 0, end: int = 1) -> Tuple:
    """
    Pick one or more element from params to be used in plot
    Args:
        data: state values at every continuation step
        start: start index
        end: end index

    Returns:
        Tuple of lists y, x
    """
    # print(data[0], data[1])
    param = []
    bparam = []
    value = []
    for k in range(len(data)):
        param.append(data[k][0][start:end])
        bparam.append(data[k][1][0])
        value.append(data[k][2][0])
    return param, bparam, value


def cosine_data_transform(data) -> Tuple:
    """
    Delta dot product between consecutive delta-param vectors to be used in plot
    Args:
        data: state values at every continuation step

    Returns:
        Tuple of lists y, x
    """
    params = []
    bparam = []
    value = []
    k = 0
    while k + 2 < len(data):
        params.append(
            pytree_dot(
                pytree_sub(data[k][0], data[k + 1][0]),
                pytree_sub(data[k + 1][0], data[k + 2][0]),
            )
        )
        bparam.append(data[k][1][0])
        value.append(data[k][2][0])
        k = k + 1
    return params, bparam, value


def norm_data_transform(data) -> Tuple:
    """
    L2 norm of param vectors to be used in plot.
    Args:
        data: state values at every continuation step

    Returns:
        Tuple of lists y, x
    """
    params = []
    bparam = []
    value = []
    for k in range(len(data)):
        params.append(l2_norm(data[k][0]))
        bparam.append(data[k][1][0])
        value.append(data[k][2][0])
    return params, bparam, value


def read_data(path):
    data = []
    with jsonlines.open(path) as reader:
        for obj in reader.iter(type=list, skip_invalid=True, skip_empty=True):
            for k in range(len(obj)):
                data.append(
                    [
                        flatten_util.ravel_pytree(obj[0])[0],
                        flatten_util.ravel_pytree(obj[1])[0],
                        flatten_util.ravel_pytree(obj[2])[0],
                    ]
                )
    return data


# TODO
def export():
    pass


# TODO
def perturbplot():
    pass


def get_loss(thetas):
    z = onp.random.normal(1.0, 1.0, len(thetas))
    return z


def bif_plot(dpath, func, n=3):
    cmaps = [
        "Greys",
        "Reds",
        "Purples",
        "Blues",
        "Greens",
        "Oranges",
        "YlOrBr",
        "YlOrRd",
        "OrRd",
        "PuRd",
        "RdPu",
        "BuPu",
        "GnBu",
        "PuBu",
        "YlGnBu",
        "PuBuGn",
        "BuGn",
        "YlGn",
    ]
    # cmaps = ['coolwarm', 'PuOr']

    for i in range(n):
        _path = f"{dpath}/version_{i}.json"
        data = read_data(_path)
        y, x, z = func(data)
        # plt.plot(x, y)
        plt.scatter(x, y, c=z, cmap=cmaps[i] + "_r", alpha=1.0)
        plt.plot(x, y, alpha=1.0)

    plt.ylabel(f"{func.__name__} Network Parameters")
    plt.xlabel(f"Continuation Parameter")

    sm = plt.cm.ScalarMappable(
        cmap=cmaps[1] + "_r", norm=plt.Normalize(vmin=min(z), vmax=max(z))
    )
    plt.colorbar(sm)
    clb = plt.colorbar(sm)
    clb.ax.set_title('Train Loss')
    plt.show()
    plt.clf()


def bif_plotv(path, func):
    cmaps = [
        "Greys",
        "Purples",
        "Blues",
        "Greens",
        "Oranges",
        "Reds",
        "YlOrBr",
        "YlOrRd",
        "OrRd",
        "PuRd",
        "RdPu",
        "BuPu",
        "GnBu",
        "PuBu",
        "YlGnBu",
        "PuBuGn",
        "BuGn",
        "YlGn",
    ]
    path = f"{path}/version.json"
    data = read_data(path)
    y, x, z = func(data)
    # plt.plot(x, y)
    plt.scatter(x, y, c=z, cmap="coolwarm_r", alpha=0.6)
    plt.ylabel(f"{func.__name__} Network Parameters")
    plt.xlabel(f"Continuation Parameter")
    sm = plt.cm.ScalarMappable(
        cmap="coolwarm_r", norm=plt.Normalize(vmin=min(z), vmax=max(z))
    )
    clb = plt.colorbar(sm)
    clb.ax.set_title('Train Loss')
    plt.show()
    plt.clf()


if __name__ == "__main__":

    path = f"/opt/ml/output/data_c/parc"
    bif_plot(path, pick_array, 1)
    # bif_plotv(path, norm_data_transform)
    #bif_plot(path, pick_array, 5)
    # bif_plotv(path, norm_data_transform)
    # bif_plot(path, cosine_data_transform, 2)
    # bif_plot(path, norm_data_transform, 2)



import cProfile
import pstats
from functools import wraps


def profile(
    output_file="/opt/ml/profile/code_prof",
    sort_by="cumulative",
    lines_to_print=None,
    strip_dirs=False,
):
    """A time profiler decorator.
    Inspired by and modified the profile decorator of Giampaolo Rodola:
    http://code.activestate.com/recipes/577817-profile-decorator/
    Args:
        output_file: str or None. Default is None
            Path of the output file. If only name of the file is given, it's
            saved in the current directory.
            If it's None, the name of the decorated function is used.
        sort_by: str or SortKey enum or tuple/list of str/SortKey enum
            Sorting criteria for the Stats object.
            For a list of valid string and SortKey refer to:
            https://docs.python.org/3/library/profile.html#pstats.Stats.sort_stats
        lines_to_print: int or None
            Number of lines to print. Default (None) is for all the lines.
            This is useful in reducing the size of the printout, especially
            that sorting by 'cumulative', the time consuming operations
            are printed toward the top of the file.
        strip_dirs: bool
            Whether to remove the leading path info from file names.
            This is also useful in reducing the size of the printout
    Returns:
        Profile of the decorated function
    """

    def inner(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            _output_file = output_file or func.__name__ + ".prof"
            pr = cProfile.Profile()
            pr.enable()
            retval = func(*args, **kwargs)
            pr.disable()
            pr.dump_stats(_output_file)

            with open(_output_file, "w") as f:
                ps = pstats.Stats(pr, stream=f)
                if strip_dirs:
                    ps.strip_dirs()
                if isinstance(sort_by, (tuple, list)):
                    ps.sort_stats(*sort_by)
                else:
                    ps.sort_stats(sort_by)
                ps.print_stats(lines_to_print)
            return retval

        return wrapper

    return inner



from cjax.utils import generate_data_01
from examples.abstract_problem import AbstractProblem
from jax.experimental.optimizers import l2_norm
from jax.tree_util import tree_map
from jax.nn import sigmoid
import jax.numpy as np

inputs, outputs, param, bparam, init_fun, predict_fun = generate_data_01()


class RandomExp(AbstractProblem):
    def __init__(self):
        pass

    @staticmethod
    def objective(params, bparam) -> float:
        logits = predict_fun(params, inputs, bparam=bparam[0], activation_func=sigmoid)
        loss = np.mean((np.subtract(logits, outputs)))
        loss += l2_norm(params) + l2_norm(bparam)
        return loss

    def initial_value(self):
        return param, bparam

    def initial_values(self):
        state, bparam = self.initial_value()
        state_1 = tree_map(lambda a: a + 0.005, state)
        states = [state, state_1]
        bparam_1 = tree_map(lambda a: a + 0.10, bparam)
        bparams = [bparam, bparam_1]
        return states, bparams



from cjax.continuation.methods.predictor.base_predictor import Predictor
from jax.tree_util import tree_multimap
from cjax.utils.math_trees import *
from jax import jit


class SecantPredictor(Predictor):
    def __init__(self, concat_states, delta_s, omega, net_spacing_param, net_spacing_bparam, hparams):
        super().__init__(concat_states)
        self._prev_state = None
        self._prev_bparam = None
        self.delta_s = delta_s
        self.omega = omega
        self.net_spacing_param = net_spacing_param
        self.net_spacing_bparam = net_spacing_bparam
        self.secantvar_state = None
        self.secantvar_bparam = None
        self.secant_direction = dict()
        self.prev_secant_direction = None
        self.hparams = hparams

    def _assign_states(self):
        self._prev_state, self._prev_bparam = self._concat_states[0]
        self._state, self._bparam = self._concat_states[1]
        self.prev_secant_direction = self._concat_states[2]

    def _choose_direction(self):
        if self.prev_secant_direction:
            inner_prod = pytree_dot(
                pytree_normalized(self.prev_secant_direction),
                pytree_normalized(self.secant_direction),
            )
            state_neg_inner_prod = pytree_dot(
                [
                    self.prev_secant_direction["state"],
                    self.prev_secant_direction["bparam"],
                ],
                [
                    pytree_element_mul(self.secant_direction["state"], -1.0),
                    self.secant_direction["bparam"],
                ],
            )
            bparam_neg_inner_prod = pytree_dot(
                [
                    self.prev_secant_direction["state"],
                    self.prev_secant_direction["bparam"],
                ],
                [
                    self.secant_direction["state"],
                    pytree_element_mul(self.secant_direction["bparam"], -1.0),
                ],
            )
            neg_inner_prod = pytree_dot(
                self.prev_secant_direction,
                pytree_element_mul(self.secant_direction, -1.0),
            )
            prods = [
                inner_prod,
                state_neg_inner_prod,
                bparam_neg_inner_prod,
                neg_inner_prod,
            ]

            # prod_diffs = list(map(lambda x: abs(x - 1.0), prods))
            # print(prod_diffs)
            min_index = prods.index(max(prods))

            if min_index == 1:
                self.secant_direction.update(
                    {"state": pytree_element_mul(self.secant_direction["state"], -1.0)}
                )
            elif min_index == 2:
                self.secant_direction.update(
                    {
                        "bparam": pytree_element_mul(
                            self.secant_direction["bparam"], -1.0
                        )
                    }
                )
            elif min_index == 3:
                self.secant_direction = pytree_element_mul(self.secant_direction, -1.0)
            else:
                pass

    @staticmethod
    @jit
    def _compute_secant(_state, _bparam, _prev_state, _prev_bparam, net_spacing_param, net_spacing_bparam, omega):
        secant_direction = {}
        state_sub = pytree_sub(_state, _prev_state)
        secant_direction.update(
            {
                "state": pytree_element_mul(
                    state_sub, net_spacing_param/(l2_norm(state_sub) + 1e-2)
                )
            }
        )
        bparam_sub = pytree_sub(_bparam, _prev_bparam)
        secant_direction.update(
            {"bparam": pytree_element_mul(bparam_sub, net_spacing_bparam/np.square(l2_norm(bparam_sub)))
             }
        )
        return secant_direction


    def prediction_step(self):
        """Given current state predict next state.
        Updates (state_guess: problem parameters, bparam_guess: continuation parameter)
        """
        self._assign_states()

        self.secant_direction = self._compute_secant(self._state, self._bparam, self._prev_state,
                                                     self._prev_bparam, self.net_spacing_param,
                                                     self.net_spacing_bparam, self.omega)
        # self._choose_direction()
        try_state = None
        try_bparam = None
        for u in range(self.hparams['retry']):
            try_state = tree_multimap(
                lambda a, b: a + self.omega * b, self._state, self.secant_direction["state"]
            )
            try_bparam = tree_multimap(
                lambda a, b: a + self.omega * b,
                self._bparam,
                self.secant_direction["bparam"],
            )
            relative_error = l2_norm(pytree_relative_error(self._bparam, try_bparam))
            if self.hparams['adaptive']:
                if (
                        relative_error > self.hparams['re_tresh']
                ):
                    self.omega = self.omega * self.hparams['omega_d']
                    print(
                        f"retry as relative_error: {relative_error}"
                    )
                else:
                    break
            else:
                break

        self._state = try_state
        self._bparam = try_bparam
        print('predictor', self._bparam)
        del try_bparam, try_state

    def get_secant_vector_concat(self):
        """Concatenated secant vector.

        Returns:
          [state_vector: problem parameters, bparam_vector: continuation parameter] list
        """
        return self.secant_direction

    # TODO: below function needed?
    def get_secant_concat(self):
        """Concatenated secant guess/point.

        Returns:
          [state_guess: problem parameters, bparam_guess: continuation parameter] list
        """
        concat = dict()
        concat.update({"state": self._state})
        concat.update({"bparam": self._bparam})
        return concat



from typing import Tuple

from jax import grad, jit

from cjax.continuation.methods.corrector.base_corrector import Corrector
from cjax.optimizer.optimizer import Optimizer


class UnconstrainedCorrector(Corrector):
    """Minimize the objective using gradient based method."""

    def __init__(
        self, optimizer: Optimizer, objective, concat_states, grad_fn, warmup_period
    ):
        self.concat_states = concat_states
        self._state = None
        self._bparam = None
        self.opt = optimizer
        self.objective = objective
        self.warmup_period = warmup_period
        self.grad_fn = grad_fn

    def _assign_states(self):
        self._state, self._bparam = self.concat_states

    def correction_step(self) -> Tuple:
        """Given the current state optimize to the correct state.

        Returns:
          (state: problem parameters, bparam: continuation parameter) Tuple
        """
        self._assign_states()
        for k in range(self.warmup_period):
            grads = self.grad_fn(self._state, self._bparam)
            self._state = self.opt.update_params(self._state, grads[0])
        return self._state, self._bparam



import jax.numpy as np
from jax import grad
from jax.config import config
from typing import Tuple
from examples.abstract_problem import AbstractProblem
from jax.tree_util import *
from jax.experimental.optimizers import l2_norm

config.update("jax_debug_nans", True)


class QuadraticProblem(AbstractProblem):
    def __init__(self):
        self.inputs = None
        self.outputs = None
        self.HPARAMS_PATH = "examples/toy/hparams.json"

    @staticmethod
    def objective(params: list, bparam: list) -> float:
        result = 0.0
        for w1 in params:
            result += np.mean(
                np.divide(np.power(w1, 2), 2.0)
                * np.square(bparam[0])
            )
        return result

    def initial_values(self) -> Tuple:
        """
        PyTreeDef(list, [PyTreeDef(tuple, [*,*])])
        :return:
        """
        states = [
            [np.array([0.05])],
            [np.array([0.03])],
        ]
        # states = [
        #     [np.array([-1.734])],
        #     [np.array([-1.632])],
        # ]
        bparams = [[np.array([3.1])], [np.array([2.8])]]

        return states, bparams

    def initial_value(self) -> Tuple:
        """
        PyTreeDef(list, [PyTreeDef(tuple, [*,*])])
        :return:
        """
        state = [np.array([0.04])]
        bparam = [np.array([3.0])]
        return state, bparam


class PitchForkProblem(AbstractProblem):
    def __init__(self):
        self.inputs = None
        self.outputs = None
        self.HPARAMS_PATH = "examples/toy/hparams.json"

    @staticmethod
    def objective(params: list, bparam: list) -> float:
        result = 25.0
        for w1 in params:
            result += np.mean(
                np.divide(np.power(w1, 4), 4.0)
                + bparam[0] * np.divide(np.power(w1, 2), 2.0)
            )
        return result

    def initial_values(self) -> Tuple:
        """
        PyTreeDef(list, [PyTreeDef(tuple, [*,*])])
        :return:
        """
        states = [
            [np.array([0.05])],
            [np.array([0.03])],
        ]
        # states = [
        #     [np.array([-1.734])],
        #     [np.array([-1.632])],
        # ]
        bparams = [[np.array([1.1])], [np.array([0.8])]]

        return states, bparams

    def initial_value(self) -> Tuple:
        """
        PyTreeDef(list, [PyTreeDef(tuple, [*,*])])
        :return:
        """
        state = [np.array([0.04])]
        bparam = [np.array([3.0])]
        return state, bparam


class VectorPitchFork(AbstractProblem):
    # TODO: check self with jax grad
    # TODO: flatten vs unflatten grad at various levels

    def __init__(self):
        """
        I can't pass param and bparam here because of grad tractibility issue
        """
        self.inputs = 0.0
        self.outputs = 0.0
        self.HPARAMS_PATH = "examples/toy/hparams.json"

    @staticmethod
    def objective(state, bparam):
        """
        Computes scalar objective.
        :param params: pytree PyTreeDef(list, [PyTreeDef(tuple, [*,*])])
        :param bparam: pytree *
        :param inputs: pytree *
        :param outputs: pytree *
        :return: pytree (scalar) *
        """
        result = 0.0

        for (w, b) in state:
            result += np.mean(
                np.sum(
                    np.divide(np.power(w, 4), 4.0)
                    + bparam[0] * np.divide(np.power(w, 2), 2.0)
                )
                + np.sum(
                    np.divide(np.power(b, 4), 4.0)
                    + bparam[0] * np.divide(np.power(b, 2), 2.0)
                )
            )
        return result

    def initial_value(self) -> Tuple:
        """
        PyTreeDef(list, [PyTreeDef(tuple, [*,*])])
        :return:
        """
        state = [(np.array([-1.734]), np.array([-1.732]))]
        bparam = [np.array([-3.0])]
        return state, bparam

    def initial_values(self):
        """
        PyTreeDef(list, [PyTreeDef(tuple, [*,*])])
        :return:
        """
        # states = [
        #     [(np.array([-1.734]), np.array([-1.734]))],
        #     [(np.array([-1.732]), np.array([-1.732]))],
        # ]

        # bparams = [[np.array([-3.0])], [np.array([-2.9])]]
        # states = [
        #     [np.array([-1.734])],
        #     [np.array([-1.632])],
        # ]

        states = [
            [(np.array([0.05]), np.array([0.05]))],
            [(np.array([0.02]), np.array([0.02]))],
        ]

        bparams = [
            [np.array([3.2])],
            [np.array([3.0])]
        ]

        return states, bparams


if __name__ == "__main__":
    s = PitchForkProblem()



import jax.numpy as np
import numpy as onp
from jax.experimental import stax
from jax.nn.initializers import zeros
from jax.nn import hard_tanh
from jax.experimental.optimizers import l2_norm
from jax.experimental.stax import Dense
import numpy.random as npr
from jax import random
from examples.abstract_problem import AbstractProblem
from jax.tree_util import tree_map
from cjax.utils import constant_2d, HomotopyDense, v_2d

batch_size = 1000
input_shape = (batch_size, 8)
step_size = 0.1
num_steps = 10
code_dim = 1
npr.seed(7)


def synth_batches():
    while True:
        images = npr.rand(*input_shape).astype("float32")
        yield images


batches = synth_batches()
inputs = next(batches)

# train_images, _, _, _ = mnist(permute_train=True)
# del _
# inputs = train_images[:batch_size]
# # print(inputs.shape)

u, s, v_t = onp.linalg.svd(inputs, full_matrices=False)
I = np.eye(v_t.shape[-1])
I_add = npr.normal(0.0, 0.002, size=I.shape)
noisy_I = I + I_add


# init_fun, predict_fun = stax.serial(
#     HomotopyDense(out_dim=512, W_init=constant_2d(v_t.T), b_init=zeros),
#     HomotopyDense(out_dim=2, W_init=constant_2d(noisy_I), b_init=zeros),
#     HomotopyDense(out_dim=4, W_init=constant_2d(noisy_I), b_init=zeros),
#     Dense(out_dim=input_shape[-1], W_init=constant_2d(v_t), b_init=zeros),
# )

init_fun, predict_fun = stax.serial(
    HomotopyDense(out_dim=4, W_init=v_2d(v_t.T), b_init=zeros),
    HomotopyDense(out_dim=2, W_init=constant_2d(noisy_I), b_init=zeros),
    HomotopyDense(out_dim=4, W_init=constant_2d(noisy_I), b_init=zeros),
    Dense(out_dim=input_shape[-1], W_init=v_2d(v_t), b_init=zeros),
)


class PCATopologyAE(AbstractProblem):
    def __init__(self):
        self.HPARAMS_PATH = "examples/autoencoder/hparams.json"

    @staticmethod
    def objective(params, bparam) -> float:
        logits = predict_fun(
            params, inputs, bparam=bparam[0], activation_func=hard_tanh
        )
        loss = np.mean(np.square((np.subtract(logits, inputs))))
        loss += 0.1 * (l2_norm(params) + l2_norm(bparam))
        return loss

    def initial_value(self):
        ae_shape, ae_params = init_fun(random.PRNGKey(0), input_shape)
        # print(ae_shape)
        assert ae_shape == input_shape
        bparam = [np.array([0.01], dtype=np.float64)]
        return ae_params, bparam

    def initial_values(self):
        state_0, bparam_0 = self.initial_value()
        state_1 = tree_map(lambda a: a - 0.005, state_0)
        states = [state_0, state_1]
        bparam_1 = tree_map(lambda a: a + 0.05, bparam_0)
        bparams = [bparam_0, bparam_1]
        return states, bparams


if __name__ == "__main__":
    problem = PCATopologyAE()
    ae_params, bparams = problem.initial_value()
    loss = problem.objective(ae_params, bparams)
    print(loss)

    # init_c = constant_2d(I)
    # print(init_c(key=0, shape=(8,8)))



from cjax.continuation.arc_len_continuation import PseudoArcLenContinuation
from cjax.continuation.states.state_variables import StateWriter
from cjax.continuation.methods.predictor.secant_predictor import SecantPredictor
from cjax.continuation.methods.corrector.perturbed_constrained_fixed_corrector import (
    PerturbedFixedCorrecter,
)
import jax.numpy as np
from jax.tree_util import *
import copy
from cjax.utils import profile
import gc
from jax.experimental.optimizers import l2_norm


# TODO: make **kwargs availible


class PerturbedPseudoArcLenFixedContinuation(PseudoArcLenContinuation):
    """Noisy Pseudo Arc-length Continuation strategy.

    Composed of secant predictor and noisy constrained corrector"""

    def __init__(
        self,
        state,
        bparam,
        state_0,
        bparam_0,
        counter,
        objective,
        dual_objective,
        hparams,
        key_state,
    ):
        super().__init__(
            state,
            bparam,
            state_0,
            bparam_0,
            counter,
            objective,
            dual_objective,
            hparams,
        )
        self.key_state = key_state
        self.clip_lambda_max = lambda g: np.where(
            (g > self.hparams["lambda_max"]), self.hparams["lambda_max"], g
        )
        self.clip_lambda_min = lambda g: np.where(
            (g < self.hparams["lambda_min"]), self.hparams["lambda_min"], g
        )

    @profile(sort_by="cumulative", lines_to_print=10, strip_dirs=True)
    def run(self):
        """Runs the continuation strategy.

        A continuation strategy that defines how predictor and corrector components of the algorithm
        interact with the states of the mathematical system.
        """
        self.sw = StateWriter(f"{self.output_file}/version_{self.key_state}.json")

        for i in range(self.continuation_steps):
            print(self._value_wrap.get_record(), self._bparam_wrap.get_record())
            self._state_wrap.counter = i
            self._bparam_wrap.counter = i
            self._value_wrap.counter = i
            self.sw.write(
                [
                    self._state_wrap.get_record(),
                    self._bparam_wrap.get_record(),
                    self._value_wrap.get_record(),
                ]
            )

            concat_states = [
                (self._prev_state, self._prev_bparam),
                (self._state_wrap.state, self._bparam_wrap.state),
                self.prev_secant_direction,
            ]

            predictor = SecantPredictor(
                concat_states=concat_states,
                delta_s=self._delta_s,
                omega=self._omega,
                net_spacing_param=self.hparams["net_spacing_param"],
                net_spacing_bparam=self.hparams["net_spacing_bparam"],
                hparams=self.hparams
            )
            predictor.prediction_step()

            self.prev_secant_direction = predictor.secant_direction

            self.hparams['sphere_radius'] = 2*self.hparams['omega']*l2_norm(predictor.secant_direction)
            concat_states = [
                predictor.state,
                predictor.bparam,
                predictor.secant_direction,
                {"state": predictor.state, "bparam": predictor.bparam},
            ]
            del predictor
            gc.collect()
            corrector = PerturbedFixedCorrecter(
                optimizer=self.opt,
                objective=self.objective,
                dual_objective=self.dual_objective,
                lagrange_multiplier=self._lagrange_multiplier,
                concat_states=concat_states,
                delta_s=self._delta_s,
                ascent_opt=self.ascent_opt,
                key_state=self.key_state,
                compute_min_grad_fn=self.compute_min_grad_fn,
                compute_max_grad_fn=self.compute_max_grad_fn,
                compute_grad_fn=self.compute_grad_fn,
                hparams=self.hparams,
                pred_state=[self._state_wrap.state, self._bparam_wrap.state],
                pred_prev_state=[self._state_wrap.state, self._bparam_wrap.state],
                counter=self.continuation_steps,
            )
            self._prev_state = copy.deepcopy(self._state_wrap.state)
            self._prev_bparam = copy.deepcopy(self._bparam_wrap.state)

            (
                state,
                bparam,
                quality
            ) = (
                corrector.correction_step()
            )  # TODO: make predictor corrector similar api's



            bparam = tree_map(self.clip_lambda_max, bparam)
            bparam = tree_map(self.clip_lambda_min, bparam)
            value = self.value_func(state, bparam)
            self._state_wrap.state = state
            self._bparam_wrap.state = bparam
            self._value_wrap.state = value

            del corrector
            del concat_states
            gc.collect()
            if (bparam[0]>=self.hparams['lambda_max']) or (bparam[0]<=self.hparams['lambda_min']):
                self.sw.write(
                    [
                        self._state_wrap.get_record(),
                        self._bparam_wrap.get_record(),
                        self._value_wrap.get_record(),
                    ]
                )
                break




from typing import Tuple
from abc import ABC, abstractmethod


class AbstractPredictor(ABC):
    """Abstract Predictor to be inherited by developer for any new predictor."""

    @abstractmethod
    def prediction_step(self):
        pass

    @abstractmethod
    def _assign_states(self):
        pass


class Predictor(AbstractPredictor):  # Todo state maintiner ?
    def __init__(self, concat_states):
        self._concat_states = concat_states
        self._state = None
        self._bparam = None

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, state):
        self._state = state

    @property
    def bparam(self):
        return self._bparam

    @bparam.setter
    def bparam(self, bparam):
        self._bparam = bparam

    def _assign_states(self):
        self._state = self._concat_states[0]
        self._bparam = self._concat_states[1]

    def prediction_step(self):
        pass



from cjax.continuation.methods.predictor.base_predictor import Predictor
from cjax.utils.math_trees import pytree_element_add


class NaturalPredictor(Predictor):
    """Natural Predictor only updates continuation parameter"""

    def __init__(self, concat_states, delta_s):
        super().__init__(concat_states)
        self.delta_s = delta_s

    def _assign_states(self) -> None:
        super()._assign_states()

    def prediction_step(self):
        """Given current state predict next state.
        Updates (state: problem parameters, bparam: continuation parameter) Tuple
        """
        self._assign_states()
        self._bparam = pytree_element_add(self._bparam, self.delta_s)
